from . import collector
from .. import stellar_info
from ..colors import TITLE, SECTION, TYPE, STAT, RESET


def new_collector():
    return ScannedBodies()


def get_description():
    return "Counts of all scanned stellar bodies"


def setup_parser(parser):
    pass


class ScannedBodies(collector.Collector):
    bodies_scanned = None
    planet_classes_scanned = None
    star_types_scanned = None
    total = 0

    def __init__(self):
        super().__init__()
        
        self.bodies_scanned = set()
        self.planet_classes_scanned = {}
        self.star_types_scanned = {}

    def process_event(self, event):
        if event["event"] != "Scan":
            return
        if "PlanetClass" not in event and "StarType" not in event:
            return
        if event["BodyName"] in self.bodies_scanned:
            return
        
        self.bodies_scanned.add(event["BodyName"])
        self.total += 1
        
        if "PlanetClass" in event:
            planet_class = event["PlanetClass"]
            
            if planet_class not in self.planet_classes_scanned:
                self.planet_classes_scanned[planet_class] = 0
            
            self.planet_classes_scanned[planet_class] += 1
        elif "StarType" in event:
            star_type = event["StarType"]
            
            if star_type not in self.star_types_scanned:
                self.star_types_scanned[star_type] = 0
            
            self.star_types_scanned[star_type] += 1

    def get_stellar_remnant_names(self):
        names = {}
    
        for type in stellar_info.stellar_remnant_types():
            names[stellar_info.type_to_name(type)] = type
        
        return names

    def get_output(self):
        self.add_line(f"{TITLE}Scanned bodies\n")
        self.add_line(f"Total: {STAT}{self.total}\n")
        
        self.add_line(f"{SECTION}Stellar remnants:")
        stellar_remnant_names = self.get_stellar_remnant_names()
        for name in sorted(stellar_remnant_names):
            type = stellar_remnant_names[name]
            count = self.star_types_scanned.get(type, 0)
            self.add_line(f"\t{TYPE}{name}{RESET}: {STAT}{count}")
            
            if type in self.star_types_scanned:
                del self.star_types_scanned[type]
        
        self.add_line(f"{SECTION}Main sequence stars:")
        for type in ["O", "B", "A", "F", "G", "K", "M"]:
            self.add_line(f"\t{TYPE}{type} star{RESET}: {STAT}{self.star_types_scanned.get(type, 0)}")
            
            if type in self.star_types_scanned:
                del self.star_types_scanned[type]
        
        self.add_line(f"{SECTION}Dwarf stars:")
        for type in ["Y", "T", "L"]:
            self.add_line(f"\t{TYPE}{type} brown dwarf{RESET}: {STAT}{self.star_types_scanned.get(type, 0)}")
            
            if type in self.star_types_scanned:
                del self.star_types_scanned[type]
        
        self.add_line(f"{SECTION}Other:")
        for type in sorted(self.star_types_scanned):
            name = stellar_info.type_to_name(type)
            self.add_line(f"\t{TYPE}{name}{RESET}: {STAT}{self.star_types_scanned[type]}")
        
        self.add_line(f"{SECTION}Planets:")
        for planet_class in sorted(self.planet_classes_scanned):
            self.add_line(f"\t{TYPE}{planet_class}{RESET}: {STAT}{self.planet_classes_scanned[planet_class]}")
        
        return self._output



from . import collector
from ..colors import TITLE, SECTION, TYPE, STAT, RESET


def new_collector():
    return Biological()


def get_description():
    return "All exobiology on-foot scans by genus and species"


def setup_parser(parser):
    pass


class Biological(collector.Collector):
    scanned_genus = None
    latest_localisation = None
    total = 0

    def __init__(self):
        super().__init__()
        
        self.scanned_genus = {}
        self.latest_localisation = {}

    def process_event(self, event):
        if event["event"] != "ScanOrganic":
            return
        if event["ScanType"] != "Analyse":
            return
        
        self.total += 1
        
        genus_id = event["Genus"]
        genus_name = event["Genus_Localised"]
        species_id = event["Species"]
        species_name = event["Species_Localised"]
        
        self.latest_localisation[genus_id] = genus_name
        self.latest_localisation[species_id] = species_name
        
        if genus_id not in self.scanned_genus:
            self.scanned_genus[genus_id] = {
                species_id: 1
            }
        
        if species_id not in self.scanned_genus[genus_id]:
            self.scanned_genus[genus_id][species_id] = 0
        
        self.scanned_genus[genus_id][species_id] += 1

    def get_output(self):
        self.add_line(f"{TITLE}Biological data\n")
        self.add_line(f"Total analysed: {STAT}{self.total}\n")
        
        for genus_id in self.scanned_genus:
            genus_total = 0
            for species_id in self.scanned_genus[genus_id]:
                genus_total += self.scanned_genus[genus_id][species_id]
            
            genus_name = self.latest_localisation[genus_id]
            self.add_line(f"{SECTION}{genus_name} genus{RESET} ({STAT}{genus_total}{RESET} total)")
            
            for species_id in self.scanned_genus[genus_id]:
                species_name = self.latest_localisation[species_id]
                self.add_line(f"\t{TYPE}{species_name}{RESET}: {STAT}{self.scanned_genus[genus_id][species_id]}{RESET} analysed")
        
        return self._output



_star_type_names = {
    "N": "Neutron star",
    "W": "Wolf-Rayet star (W)",
    "WN": "Wolf-Rayet star (WN)",
    "WNC": "Wolf-Rayet star (WNC)",
    "WC": "Wolf-Rayet star (WC)",
    "WO": "Wolf-Rayet star (WO)",
    "TTS": "T Tauri star",
    "AeBe": "Herbig Ae/Be star",
    "T": "T brown dwarf",
    "L": "L brown dwarf",
    "Y": "Y brown dwarf",
    "O": "O star",
    "B": "B star",
    "B_BlueWhiteSuperGiant": "B blue-white supergiant star",
    "A": "A star",
    "F": "F star",
    "F_WhiteSuperGiant": "F white supergiant star",
    "G": "G star",
    "G_WhiteSuperGiant": "G white supergiant star",
    "K": "K star",
    "K_OrangeGiant": "K orange giant star",
    "M": "M star",
    "M_RedGiant": "M red giant star",
    "M_RedSuperGiant": "M red supergiant star",
    "S": "S star",
    "MS": "MS star",
    "C": "Carbon star (C)",
    "CS": "Carbon star (C)",
    "CN": "Carbon star (C)",
    "CJ": "Carbon star (C)",
    "CH": "Carbon star (C)",
    "CHd": "Carbon star (C)",
    "H": "Black hole",
    "SupermassiveBlackHole": "Supermassive black hole",
    "D":    "White dwarf (D)",
    "DA":   "White dwarf (DA)",
    "DAB":  "White dwarf (DAB)",
    "DAO":  "White dwarf (DAO)",
    "DAZ":  "White dwarf (DAZ)",
    "DAV":  "White dwarf (DAV)",
    "DB":   "White dwarf (DB)",
    "DBZ":  "White dwarf (DBZ)",
    "DBV":  "White dwarf (DBV)",
    "DO":   "White dwarf (DO)",
    "DOV":  "White dwarf (DOV)",
    "DQ":   "White dwarf (DQ)",
    "DC":   "White dwarf (DC)",
    "DCV":  "White dwarf (DCV)",
    "DX":   "White dwarf (DX)",
}


def type_to_name(type: str) -> str:
    if not type:
        return f"Invalid type: {type}"

    if type in _star_type_names:
        return _star_type_names[type]
    
    return f"Unknown type '{type}'"


def sorted_types() -> list[str]:
    return [
        "O",
        "B", "B_BlueWhiteSuperGiant",
        "A",
        "F", "F_WhiteSuperGiant",
        "G", "G_WhiteSuperGiant",
        "K", "K_OrangeGiant",
        "M", "M_RedGiant", "M_RedSuperGiant",
        "T", "L", "Y",
        "N",
        "H", "SupermassiveBlackHole",
        "D", "DA", "DAB", "DAO", "DAZ", "DAV", "DB", "DBZ", "DBV", "DC", "DCV", "DO", "DOV", "DQ", "DX",
        "W", "WN", "WNC", "WC", "WO",
        "TTS",
        "AeBe",
        "S", "MS",
        "C", "CS", "CN", "CJ", "CH", "CHd",
    ]


def star_types() -> list[str]:
    return list(_star_type_names.keys())


def stellar_remnant_types() -> set[str]:
    types = set()
    white_dwarf_types = ["D", "DA", "DAB", "DAO", "DAZ", "DAV", "DB", "DBZ", "DBV",
                         "DO", "DOV", "DQ", "DC", "DCV", "DX"]
    
    types.update(white_dwarf_types)
    types.add("H")
    types.add("N")
    types.add("SupermassiveBlackHole")
    
    return types



from . import collector
from ..colors import TITLE, STAT, RESET


def new_collector():
    return MostBodies()


def get_description():
    return "System with the most stellar objects"


def setup_parser(parser):
    pass


class MostBodies(collector.Collector):
    best_system_name = None
    best_body_count = None

    def __init__(self):
        super().__init__()
        
        self.best_body_count = 0

    def process_event(self, event):
        if event["event"] != "FSSDiscoveryScan": # Relying entirely on honking events can miss manual discovery
            return
        
        count = event["BodyCount"]
        system = event["SystemName"]
        
        if count > self.best_body_count:
            self.best_system_name = system
            self.best_body_count = count

    def get_output(self):
        self.add_line(f"{TITLE}Most bodies{RESET}: {STAT}{self.best_body_count}{RESET} at system {self.best_system_name}")
        
        return self._output



from . import min_max_collector
from .. import stellar_info
from .. import time_formatting
from ..colors import TITLE, SECTION


def new_collector():
    return RotationPeriod()


def get_description():
    return "Objects with notable rotational periods"


def setup_parser(parser):
    pass


class RotationPeriod(min_max_collector.MinMaxCollector):
    def __init__(self):
        super().__init__()

    def process_event(self, event):
        if event["event"] != "Scan":
            return
        if "RotationPeriod" not in event:
            return
    
        object_info = self.get_object_info(event["BodyName"], event["StarSystem"], abs(event["RotationPeriod"]))
        
        if "StarType" in event:
            self.check_body(self.notable_stars, event["StarType"], object_info)
        elif "PlanetClass" in event:
            self.check_body(self.notable_bodies, event["PlanetClass"], object_info)

    def get_output(self):
        self.add_line(f"{TITLE}Rotation period\n")
        
        self.add_line(f"{SECTION}Stars:")
        for type in stellar_info.sorted_types():
            if type not in self.notable_stars:
                continue
            
            self.add_type_info(stellar_info.type_to_name(type), self.notable_stars[type])
            self.add_line()
        
        self.add_line()
        self.add_line(f"{SECTION}Planets/moons:")
        for type in sorted(self.notable_bodies):
            self.add_type_info(type, self.notable_bodies[type])
            self.add_line()
        
        return self._output

    def get_formatted_stat(self, stat):
        return time_formatting.format_period(stat)



import datetime

MAX_UNITS = 2


def format_period(seconds):
    delta = datetime.timedelta(seconds=seconds)
    
    microseconds = delta.microseconds % 1000
    milliseconds = delta.microseconds // 1000
    seconds = delta.seconds % 60
    minutes = delta.seconds // 60 % 60
    hours = delta.seconds // (60 * 60)
    
    days = delta.days % 365
    years = delta.days // 365
    
    units = [f"{microseconds} microsecond{_get_plural(microseconds)}"]
    
    if milliseconds > 0:
        units.append(f"{milliseconds} millisecond{_get_plural(milliseconds)}")
    if seconds > 0:
        units.append(f"{seconds} second{_get_plural(seconds)}")
    if minutes > 0:
        units.append(f"{minutes} minute{_get_plural(minutes)}")
    if hours > 0:
        units.append(f"{hours} hour{_get_plural(hours)}")
    if days > 0:
        units.append(f"{days} day{_get_plural(days)}")
    if years > 0:
        units.append(f"{years} year{_get_plural(years)}")
    
    units.reverse()
    
    if len(units) > MAX_UNITS:
        units = units[0:MAX_UNITS]
    
    return ", ".join(units)
    

def _get_plural(count):
    return 's' if count != 1 else ''



from . import min_max_collector
from ..colors import TITLE


def new_collector():
    return Gravity()


def get_description():
    return "The landable planets/moons with the lowest and highest gravity"


def setup_parser(parser):
    pass


class Gravity(min_max_collector.MinMaxCollector):
    def __init__(self):
        super().__init__()

    def process_event(self, event):
        if event["event"] != "Scan":
            return
        if "SurfaceGravity" not in event:
            return
        if "Landable" not in event or not event["Landable"]:
            return
        
        object_info = self.get_object_info(event["BodyName"], event["StarSystem"], event["SurfaceGravity"])
        
        self.check_body(self.notable_bodies, event["PlanetClass"], object_info)

    def get_output(self):
        self.add_line(f"{TITLE}Landable gravity\n")

        for type in sorted(self.notable_bodies):
            self.add_type_info(type, self.notable_bodies[type])
            self.add_line()
        
        return self._output

    def get_formatted_stat(self, stat):
        return f"{round(stat / 9.8, 3)}g"



from . import collector
from ..colors import TITLE, SECTION, STAT, RESET


def new_collector():
    return SurfaceScans()


def get_description():
    return "DSS probes used and efficiency"


def setup_parser(parser):
    pass


class SurfaceScans(collector.Collector):
    total_probes = 0
    total_bodies = 0
    total_rings = 0
    average_efficiency = 0

    def __init__(self):
        super().__init__()

    def process_event(self, event):
        if event["event"] != "SAAScanComplete":
            return
        
        probes_used = event["ProbesUsed"]
        efficiency_target = event["EfficiencyTarget"]
        
        self.total_probes += probes_used
        
        if efficiency_target != 0:
            self.total_bodies += 1
            self.average_efficiency += probes_used / efficiency_target
        else:
            self.total_rings += 1

    def get_output(self):
        self.add_line(f"{TITLE}Surface scans\n")
        
        self.add_line(f"{SECTION}Total probes hit{RESET}: {STAT}{self.total_probes} probes")
        self.add_line(f"{SECTION}Total bodies scanned{RESET}: {STAT}{self.total_bodies}{RESET}")
        self.add_line(f"{SECTION}Total rings scanned{RESET}: {STAT}{self.total_rings}{RESET}")
        self.add_line(f"{SECTION}Average efficiency{RESET}: {STAT}{round(self.average_efficiency / self.total_bodies * 100)}%{RESET} of target efficiency")
        
        return self._output



from . import collector
from ..colors import TITLE, STAT

def new_collector():
    return VisitedSystems()


def get_description():
    return "Total systems visited"


def setup_parser(parser):
    pass


class VisitedSystems(collector.Collector):
    visited_systems = None

    def __init__(self):
        super().__init__()
        
        self.visited_systems = set()

    def process_event(self, event):
        if event["event"] != "Scan" and event["event"] != "Location" and event["event"] != "FSDJump":
            return
            
        self.visited_systems.add(event["StarSystem"])

    def get_output(self):
        self.add_line(f"{TITLE}Visited systems")
        self.add_line(f"\tTotal visited systems: {STAT}{len(self.visited_systems)}")
        
        return self._output
    


from ..colors import RESET


class Collector:
    _output = None

    def __init__(self):
        self._output = ""

    def process_event(self, event):
        pass

    def add_line(self, text: str = ""):
        self._output += f"{text}\n{RESET}"

    def get_output(self):
        pass



import json
import os


def read_events(path, exclude_legacy=False):
    journal_paths = []

    for entry in os.listdir(path):
        entry_path = os.path.join(path, entry)
        
        if os.path.isfile(entry_path) and entry.startswith("Journal.") and entry.endswith(".log"):
            if exclude_legacy and "-" not in entry:
                continue

            journal_paths.append(entry_path)
    
    for i in range(len(journal_paths)):
        path = journal_paths[i]
        
        for event in _read_journal(path):
            yield event
        
        progress_text = f"{i+1}/{len(journal_paths)} journal files processed"
        if i == len(journal_paths) - 1:
            print(" " * len(progress_text), flush=True, end="\r")
        else:
            print(progress_text, flush=True, end="\r")


def _read_journal(path):
    with open(path) as journal_file:
        for line in journal_file:
            event = json.loads(line)
            
            yield event



import argparse
import os

import colorama

from . import read_journals
from . import stat_groups
from .stat_groups import *
from .colors import TITLE, STAT, RESET

PROGRAM_NAME = "explorer-stats"
VERSION = "1.0.1"


def main():
    colorama.init(autoreset=True)

    parser = build_arg_parser()
    
    args = parser.parse_args()
    
    if not args.stat_group:
        run_main()
    else:
        stat_group_name = args.stat_group
        if stat_group_name not in stat_groups.get_module_names():
            invalid_stat_group(stat_group_name)
        else:
            run_stat_group(args)


def run_stat_group(args):
    module = stat_groups.get_module(args.stat_group)
    collector = module.new_collector()
    
    saves_path = os.path.expandvars(args.saves_path)

    for event in read_journals.read_events(saves_path, args.exclude_legacy):
        collector.process_event(event)
    
    print(collector.get_output())


def invalid_stat_group(name):
    print(f"Invalid stat group '{name}', probably missing in package init file")


def run_main():
    print_stat_groups()


def print_stat_groups():
    for name in stat_groups.get_module_names(): 
        print(f"{TITLE}{name}{RESET} - {STAT}{stat_groups.get_module(name).get_description()}")


def build_arg_parser():
    parser = argparse.ArgumentParser(prog=PROGRAM_NAME)
    parser.add_argument("--saves_path", type=str,
                        default="%USERPROFILE%/Saved Games/Frontier Developments/Elite Dangerous",
                        help="path to the ED saved data")
    parser.add_argument("--version", action="version", version=f"{PROGRAM_NAME} v{VERSION}")
    parser.add_argument("-xl", "--exclude-legacy", "--no-legacy", action="store_true",
                        help="Exclude legacy journal files from stats", dest="exclude_legacy")
    
    subparsers = parser.add_subparsers(dest="stat_group")
    
    for name in stat_groups.get_module_names():
        subparser = subparsers.add_parser(name)
        
        stat_groups.get_module(name).setup_parser(subparser)
    
    return parser

    
if __name__ == "__main__":
    main()



from . import collector
from .. import time_formatting
from ..colors import TITLE, SECTION, STAT, RESET


def new_collector():
    return FuelScooped()


def get_description():
    return "Total fuel scooped and time spent scooping"


def setup_parser(parser):
    pass


# array of arrays, index corresponds to fuel scoop class which has an array from rating E-A.
# scoop rates are in tonnes per second
_scoop_rates = [
    [0.018, 0.024, 0.030, 0.036, 0.042],
    [0.032, 0.043, 0.054, 0.065, 0.075],
    [0.075, 0.100, 0.126, 0.151, 0.176],
    [0.147, 0.196, 0.245, 0.294, 0.343],
    [0.247, 0.330, 0.412, 0.494, 0.577],
    [0.376, 0.502, 0.627, 0.752, 0.878],
    [0.534, 0.712, 0.890, 1.068, 1.245],
    [0.720, 0.960, 1.200, 1.440, 1.680],
]


class FuelScooped(collector.Collector):
    last_fuelscoop_rate = None
    total_fuel_scooped = 0
    time_scooping = 0

    def __init__(self):
        super().__init__()

    def process_event(self, event):
        if event["event"] == "Loadout":
            for module in event["Modules"]:
                item_name = module["Item"]
                
                if "fuelscoop" in item_name:
                    self.last_fuelscoop_rate = self.get_scoop_rate(item_name)
                    break
            
            return
        
        if event["event"] != "FuelScoop":
            return
        
        fuel_scooped = event["Scooped"]
        
        self.total_fuel_scooped += fuel_scooped
        self.time_scooping += fuel_scooped / self.last_fuelscoop_rate

    def get_output(self):
        self.add_line(f"{TITLE}Fuel scooped\n")
        
        self.add_line(f"{SECTION}Total fuel scooped{RESET}: {STAT}{round(self.total_fuel_scooped)} tonnes")
        self.add_line(f"{SECTION}Time spent scooping{RESET}: {STAT}{time_formatting.format_period(self.time_scooping)}")
        
        return self._output

    def get_scoop_rate(self, full_name):
        size = int(full_name[len("int_fuelscoop_size")])
        class_rating = int(full_name[-1])
        
        return _scoop_rates[size - 1][class_rating - 1]



__all__ = [
    "scanned_bodies",
    "visited_systems",
    "most_bodies",
    "orbital_period",
    "rotation_period",
    "radius",
    "gravity",
    "fuel_scooped",
    "biological",
    "surface_scans",
]


def get_module_names():
    return __all__


def get_module(name):
    return globals()[name]



from . import collector
from ..colors import TYPE, STAT, RESET


class MinMaxCollector(collector.Collector):
    notable_stars = None
    notable_bodies = None
    
    def __init__(self):
        super().__init__()
    
        self.notable_stars = {}
        self.notable_bodies = {}

    def get_formatted_stat(self, stat):
        return stat

    def add_type_info(self, type_name, info):
        self.add_line(str(TYPE) + type_name)
        
        highest_info = info["highest"]
        lowest_info = info["lowest"]
            
        highest_system = highest_info["system"]
        lowest_system = lowest_info["system"]
        
        highest_object = highest_info["name"]
        lowest_object = lowest_info["name"]
        
        highest_formatted = self.get_formatted_stat(highest_info["stat"])
        lowest_formatted = self.get_formatted_stat(lowest_info["stat"])
        
        if highest_info == lowest_info:
            self.add_line(f"\tHighest/lowest: {STAT}{highest_formatted}{RESET} (object {highest_object} in system {highest_system})")
        else:
            self.add_line(f"\tHighest: {STAT}{highest_formatted}{RESET} (object {highest_object} in system {highest_system})")
            self.add_line(f"\tLowest: {STAT}{lowest_formatted}{RESET} (object {lowest_object} in system {lowest_system})")

    def shorten_body_name(self, name, system):
        if name == system:
            return name

        if name.startswith(system):
            return name[len(system) + 1:]

        return name

    def get_object_info(self, object_name, system_name, stat):
        return {
            "name": self.shorten_body_name(object_name, system_name),
            "stat": stat,
            "system": system_name,
        }

    def check_body(self, lookup_dict, type, object_info):
        if type not in lookup_dict:
            lookup_dict[type] = {
                "highest": object_info,
                "lowest": object_info,
            }
        else:
            if lookup_dict[type]["highest"]["stat"] < object_info["stat"]:
                lookup_dict[type]["highest"] = object_info
            if lookup_dict[type]["lowest"]["stat"] > object_info["stat"]:
                lookup_dict[type]["lowest"] = object_info
